from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from keras_applications import xception

Xception = xception.Xception
decode_predictions = xception.decode_predictions
preprocess_input = xception.preprocess_input
